/*
 * Copyright 2012, the mGui Authors. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *  1. Redistributions of source code must retain the above copyright notice and this
 *     list of conditions.
 *  2. Redistributions in binary form must reproduce the above copyright notice and this
 *     list of conditions in the documentation and/or other materials provided with the
 *     distribution.
 */

// FIXME: All controls should always be subclassed, we need a default implementation
//   to catch/bubble dropFiles handlers

#pragma once
#include <string>
#include <vector>

// MG HELPERS
// ``````````

#define MGCT_UNKNOWN  0
#define MGCT_WINDOW   1
#define MGCT_MENUITEM 2
#define MGCT_EDIT     3
#define MGCT_BUTTON   4
#define MGCT_LISTBOX  5
#define MGCT_PROGRESSBAR 6
#define MGCT_LABEL    7
#define MGCT_GROUP    8
#define MGCT_COMBOBOX 9
#define MGCT_TIMER    10
#define MGCT_HOTKEY   11
#define MGCT_HOTKEYCONTROL 12

#define MGCT_CHECKBOX 20
#define MGCT_IMAGE 21

// MGVIRTUALKEY
// ````````````
enum MGVirtualKeys {
	MGVK_NONE = 0,  // Invalid or not set

	MGVK_TAB, MGVK_ENTER, MGVK_ESC, MGVK_SPACE,	MGVK_INS, MGVK_TILDE, MGVK_DEL,

	MGVK_0,	MGVK_1, MGVK_2, MGVK_3, MGVK_4, MGVK_5, MGVK_6, MGVK_7, MGVK_8, MGVK_9,
	MGVK_A, MGVK_B, MGVK_C, MGVK_D, MGVK_E, MGVK_F, MGVK_G, MGVK_H, MGVK_I, MGVK_J,
	MGVK_K, MGVK_L, MGVK_M, MGVK_N, MGVK_O, MGVK_P, MGVK_Q, MGVK_R, MGVK_S,	MGVK_T,
	MGVK_U, MGVK_V, MGVK_W, MGVK_X, MGVK_Y, MGVK_Z,

	MGVK_F1, MGVK_F2, MGVK_F3, MGVK_F4, MGVK_F5, MGVK_F6, MGVK_F7, MGVK_F8, MGVK_F9,
	MGVK_F10, MGVK_F11, MGVK_F12,

	// (css order)
	MGVK_UP, MGVK_RIGHT, MGVK_DOWN, MGVK_LEFT,

	MGVK_MAX_VIRTKEYS
};

typedef unsigned char MGObjectType;

struct MGKeyCombination {
	enum MGVirtualKeys virtkey; // MGVK_*
	bool alt;
	bool ctrl;
	bool shift;
};

inline MGKeyCombination MGKC(enum MGVirtualKeys virtkey, bool alt, bool ctrl, bool shift) {
	MGKeyCombination ret = {virtkey, alt, ctrl, shift};
	return ret;
};

std::string MGVK_Format(enum MGVirtualKeys mgvk);
std::string MGKC_FormatCharacter(MGKeyCombination& kc);
std::string MGKC_Format(MGKeyCombination& kc);
std::string MGKC_FormatSmall(MGKeyCombination& kc);

// MISC
// ````

struct MGWindowData;
struct MGControlData;
struct MGEventData;
class  MGEventHandler;
class  MGFontManager;
class  MGMenuItem;
struct MGMenuItemData;
struct MGMenuItemRelation;
class  MGWindow;
struct MGActionData;
//class  MGAction;
struct MGListBoxData;

typedef void(MGWindow::*MGCallback)(void);
#define MGCB(x) static_cast<MGCallback>(&x)

class MGObject {
public:
	MGObjectType _controlType;

protected:
	MGObject() {}; // provide trivial constructor since noncopyable breaks it
	~MGObject() {}; // noncopyable
private:
	MGObject(const MGObject& mgo); // noncopyable
	MGObject& operator= (const MGObject& mgo); // noncopyable

};

// MGACTION
// ````````
#define MGACTION_UNUSED char
template<typename T> class MGAction {
private:
	MGWindow* pWnd;
	std::vector<MGCallback> fns;
	bool m_defaultPrevented;

public:
	T data;

	MGAction() {};
	~MGAction() {};

	// Add an additional handler to this event
	void assign(MGWindow* object, MGCallback function) {
		//if (object && function) {
		pWnd = object;
		fns.push_back(function);
		//}
	}

	void invoke() {
		m_defaultPrevented = false;
		size_t i = fns.size();
		while(i --> 0) {
			//if (! fns[i]) continue; // FIXME: why was null ever pushed..?
			// More-recently-added handlers can prevent underlying ones from occurring
			( pWnd ->*(fns[i]) )();
			if (wasEventDefaultPrevented()) break;
		}
	}

	bool isAssigned() {
		return (fns.size() > 0);
	}

	void clearAllHandlers() {
		fns.clear();
	}

	void event_preventDefault() {
		m_defaultPrevented = true;
	}

	bool wasEventDefaultPrevented() {
		return m_defaultPrevented;
	}
};

// MGCONTROL
// `````````
class MGControl : public MGObject {
protected:
	MGControlData *d;
	MGControl();
	~MGControl();

public:
	void resize(int x, int y, int width, int height);
	void resize(int width, int height);
	void move(int x, int y);
	void snapToScreenEdge();
	void refresh();
	int  getX();
	int  getY();
	int  getWidth();
	int  getHeight();
	int  getInnerWidth();
	int  getInnerHeight();

	virtual void focus();

	void setEnabled(bool bEnabled);
	void setVisible(bool bVisible);
	void removeStyles();
	void removeBorder();
	virtual long _parseEvent(MGEventData *ev);

	MGAction<std::vector<std::string>> onDropFiles;
	void refreshHasDropFilesHandler();
};


// MGWINDOW
// ````````
class MGWindow : public MGControl {
private:
	MGEventHandler *e;
	MGWindowData *w;

public:
	// Constructor: hIcon is platform-specific.
    //     MSW:  null, or (void*)(an embedded resource ID)
	//     FLTK: null, or a const pointer to 16x16 bitmap data (fixme?).
	MGWindow(void* hIcon, const char* caption, MGWindow* modalParent);
	~MGWindow();
	void _prepareChild(MGControlData *pCtrlData, MGControl *pCtrl);
	void _prepareChild(MGMenuItemData *pMenuData, MGMenuItem *pMenu);
	long _parseEvent(MGEventData *ev);
	
	bool doEvents();
	void doEventLoop();
	//void doChildModal(MGWindow* child);
	bool isActive();
	void close();
	void focus();
	
	void setResizable(bool resizable);
	void showStatusBar(bool show);
	void setStatusBar(const char* text);
	void setCaption(const char* caption);

	void addChildMenu(MGMenuItem* menuItem);
	void updateMenu();	
	
	void dlgAlert(const char* message);
	void dlgAlert(const char* message, const char* title);
	bool dlgConfirm(const char* prompt);
	bool dlgFileOpen(const char* title, const char* filter, std::string& result);
	bool dlgFolderSelect(const char* title, std::string& result);
	bool dlgFolderSelectEx(const char* title, const char* optname, std::string &result, bool &optresult);

	int getInnerHeight();

	void registerMarshall(unsigned int event_id, MGAction<long> *doCall);
	bool doMarshalled(unsigned int event_id, long data);

	MGAction<MGACTION_UNUSED> onResize;
	MGAction<MGACTION_UNUSED> onMove;
	MGAction<MGKeyCombination> onKey;
	MGAction<MGACTION_UNUSED> onClosing; // Called just before your window exits
};

// MGMENUITEM
// ``````````
// Call MGWindow::updateMenus() to reflect changes
class MGMenuItem : public MGObject {
private:
	MGMenuItemData* m;
	void _initialise(MGWindow* parent, const char* text);

public:
	MGMenuItem(MGWindow*   parent, const char* text);
	MGMenuItem(MGMenuItem* parent, const char* text);
	MGMenuItem(MGWindow*   parent, const char* text, MGCallback onClick);	
	MGMenuItem(MGMenuItem* parent, const char* text, MGCallback onClick);
	MGMenuItem(MGMenuItem* parent, const char* text, MGCallback onClick, MGKeyCombination shortcut);
	
	~MGMenuItem();
	void _updateSelf(MGMenuItemRelation* relation);

	std::string getText();
	void setText(std::string& sz);

	// Chainable methods

	// setChecked, setEnabled require the menu to be created to work; so call
	// them after MGWindow::updateMenu() has been called for the first time
	MGMenuItem* setChecked(bool isChecked);
	MGMenuItem* setEnabled(bool isEnabled);

	// Shortcut keys and child menus are created on MGWindow::updateMenu()
	MGMenuItem* setShortcut(MGKeyCombination kc);	
	MGMenuItem* addChildMenu(MGMenuItem* menuItem);
	MGMenuItem* addChildMenuSeparator();

	MGAction<MGACTION_UNUSED> onClick;
};

// MGEDIT
// ``````
class MGEdit : public MGControl {
private:
	MGEventHandler *e;

public:
	MGEdit(MGWindow* parent, bool bMultiline, bool isPassword);
	~MGEdit();
	long _parseEvent(MGEventData *ev);

	void textSet(const char* szText);
	void textAppend(const char* szText);
	std::string textGet(); //const char* outBuff, int charcount);
	int  textLength();
	void selRange(int start, int end);
	void selEnd();
		
	MGAction<MGKeyCombination> onKey;
};

// MGBUTTON
// ````````
class MGButton : public MGControl {
public:	
	MGButton(MGWindow* parent, const char* szCaption);

	MGAction<MGACTION_UNUSED> onClick;
};

class MGCheckBox : public MGControl {
public:	
	MGCheckBox(MGWindow* parent, const char* szCaption);
	bool isChecked();
	void setChecked(bool checked);

	MGAction<MGACTION_UNUSED> onCheck;
};

// MGLISTBOX
// `````````
#define NOTHING_SELECTED -1
class MGListBox : public MGControl {
private:
	MGListBoxData *z;
	MGEventHandler *e;
public:
	MGListBox(MGWindow* parent, bool bAutoSort, bool bAllowColors, int lineheight, bool bAllowMultiSelect);
	~MGListBox();
	long _parseEvent(MGEventData *ev);

	unsigned int   addItem(const char* szItem);
	unsigned int   addItem(const char* szItem, unsigned int pos);
	unsigned int   addItem_UTF16LE(const wchar_t* wzItem); // Special methods for higher win32 performance
	unsigned int   addItem_UTF16LE(const wchar_t* wzItem, unsigned int pos);

	void  clear();
	unsigned int   findItem(const char* szItem); // index or -1 on error
	unsigned int   getSelected(); // Returns where the caret is
	void  getAllSelected(std::vector<unsigned int>& idxList);
	std::string getItemText(unsigned int idx);
	unsigned int getSelectedCount();
	unsigned int   getItemCount();
	void* getItemData(unsigned int idx);
	void  setItemData(unsigned int idx, void* data);
	void  setItemColor(unsigned int idx);
	void  setItemColor(unsigned int idx, unsigned char r, unsigned char b, unsigned char g);
	void  setItemText(unsigned int idx, const char* szItem);
	void  setSelected(unsigned int idx); // index or -1 for no selection
	void  addSelection(unsigned int idx);
	void  removeItem(unsigned int idx);

	void multi_setSelected(unsigned int idx, bool isSelected); // could be better integrated..

	void setAllowTypeToSelect(bool allowTypeToSelect);

	MGAction<MGACTION_UNUSED> onChange;
	MGAction<MGKeyCombination> onKey;
	MGAction<MGACTION_UNUSED> onDoubleClick;
};

// MGPROGRESSBAR
// `````````````
class MGProgressBar : public MGControl {
private:
	MGEventHandler *e;
public:
	MGProgressBar(MGWindow* parent, bool isHorizontal, bool isSmooth);
	~MGProgressBar();
	long _parseEvent(MGEventData *ev);

	void setValue(float fraction);
	float getValue();

	MGAction<long> onMouseDown; // FIXME: use a coordinate struct!
	MGAction<long> onDrag;
};

// MGLABEL
// ```````
class MGLabel : public MGControl {
public:
	MGLabel(MGWindow* parent, bool alignLeft);

	void setText(const char* text);
	std::string getText();
};

class MGImage : public MGControl {
public:
	MGImage(MGWindow* parent, int resource_id);
};

// MGGROUP
// ```````
class MGGroup : public MGControl {
public:
	MGGroup(MGWindow* parent);

	void setText(const char* text);
};

// MGCOMBOBOX
// ``````````
class MGComboBox : public MGControl {
public:
	MGComboBox(MGWindow* parent, bool bAutoSort);
	
	void allowTextEntry(bool bEnable);
	void addItem(const char* text);
	void clear();
	int  getSelected(); // index, or -1 for error
	void setSelected(int idx);
	std::string getSelectedText();

	MGAction<MGACTION_UNUSED> onChange;
};

// MGTIMER
// ```````
class MGTimer : private MGControl {
public:
	MGTimer(MGWindow* parent, int interval_ms);
	~MGTimer();

	MGAction<MGACTION_UNUSED> onTick;
};

// MGHOTKEY
// ````````
// Systemwide hotkeys. For application-only hotkeys, use menuitem shortcuts
class MGHotkey : private MGControl {
private:
	MGKeyCombination m_kc;
public:
	MGHotkey(MGWindow* parent, MGKeyCombination kc);
	~MGHotkey();

	MGKeyCombination getCombination();
	void setCombination(MGKeyCombination kc);

	MGAction<MGACTION_UNUSED> onPress;
};

// MGHOTKEYCONTROL
// ```````````````
class MGHotkeyControl : public MGControl {
public:
	MGHotkeyControl(MGWindow* parent);

	void setValue(MGKeyCombination val);	
	MGKeyCombination getValue();
};

