/*
 * Copyright 2012, the mGui Authors. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *  1. Redistributions of source code must retain the above copyright notice and this
 *     list of conditions.
 *  2. Redistributions in binary form must reproduce the above copyright notice and this
 *     list of conditions in the documentation and/or other materials provided with the
 *     distribution.
 */

#include "mGui.h"
#define MGUI_WIN32_INTERNAL_USE
#include "mGui_Win32.h"

#include <map>


struct MGWindowData {
	bool					bActive;
	HMENU					hMenu;
	HWND					hStatus;
	bool					bStatusBarVisible;
	std::vector<MGMenuItem*> childMenus;
	std::vector<ACCEL>		childMenuAccels;
	HACCEL					hAccelTable;
	std::map<unsigned int, MGObject*> controlIDs;
	MGWindow*               modalParent;

	std::map<unsigned int, MGAction<long>*> events;
};

struct MGMenuItemData {
	std::vector<MGMenuItem*>	children;
	std::wstring				caption;
	HMENU						hMenu;
	unsigned int				ID;
	MGKeyCombination			shortcut;
	MGWindow*                   parentWnd;
	HWND                        parentHWnd;
};

struct MGMenuItemRelation {
	HMENU				parent;
	std::vector<ACCEL> *menuAccels;
};

struct MGListBoxData {
	std::map<unsigned int, unsigned int> colors;
	unsigned int default_color;
	int line_count;
	bool is_multiselect;
	bool allow_type_to_select;
};

struct MGEventData {
	HWND   hwnd;
	UINT   uMsg;
	WPARAM wParam;
	LPARAM lParam;
};

class MGEventHandler {
private:
	MGControl *m_master;
	LONG       m_oldProc;
	HWND       m_hWnd;

public:
	MGEventHandler(MGControl *master);
	~MGEventHandler();
	void subclassWnd(HWND hWnd);

	LRESULT CALLBACK Proc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	static LRESULT CALLBACK stProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);	
};


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGVirtualKey
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// REFERENCE: http://msdn.microsoft.com/en-us/library/windows/desktop/dd375731(v=vs.85).aspx

#define MAX_NATIVE_KEYCODES 0xFF
unsigned int MGVKToNative(enum MGVirtualKeys mgvk) {
	static unsigned int mgvk_lookup[MGVK_MAX_VIRTKEYS] = {
		// Unused
		0,
		// Misc. symbols
		0x09, 0x0D, 0x18, 0x20, 0x2D, 0xC0, 0x2E,
		// Numbers 0..9
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
		// Alphabet A..Z
		0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A,
		0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52, 0x53, 0x54,
		0x55, 0x56, 0x57, 0x58, 0x59, 0x5A,
		// F-keys F1..F12
		0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
		0x7A, 0x7B
	};
	if (mgvk > 0 && mgvk < MGVK_MAX_VIRTKEYS) return mgvk_lookup[mgvk];
	return 0;
};

// Returns MGVK_NONE on failure
enum MGVirtualKeys MGNativeToVK(unsigned int nativekey) {
	static bool tbInitialised = false;
	static enum MGVirtualKeys lookup[MAX_NATIVE_KEYCODES] = {MGVK_NONE}; // invalid
	if (! tbInitialised) {
		lookup[0x09     ] = MGVK_TAB;
		lookup[0x0D     ] = MGVK_ENTER;
		lookup[VK_ESCAPE] = MGVK_ESC;
		lookup[VK_SPACE ] = MGVK_SPACE;
		lookup[VK_INSERT] = MGVK_INS;
		lookup[0xC0     ] = MGVK_TILDE;
		lookup[VK_UP    ] = MGVK_UP;
		lookup[VK_RIGHT ] = MGVK_RIGHT;
		lookup[VK_DOWN  ] = MGVK_DOWN;
		lookup[VK_LEFT  ] = MGVK_LEFT;
		lookup[VK_DELETE] = MGVK_DEL;
		// Numbers 0..9
		for (int i = 0; i < 10; i++) {
			lookup[0x30 + i] = static_cast<MGVirtualKeys>(MGVK_0 + i);
		}
		// Alphabet A..Z
		for (int i = 0; i < 26; i++) {
			lookup[0x41 + i] = static_cast<MGVirtualKeys>(MGVK_A + i);
		}
		// F-keys F1..F12
		for (int i = 0; i < 12; i++) {
			lookup[0x70 + i] = static_cast<MGVirtualKeys>(MGVK_F1 + i);
		}

		// Mark table as built
		tbInitialised = true;
	}

	if (nativekey > 0 && nativekey < MAX_NATIVE_KEYCODES) {
		return lookup[nativekey];
	}
	return MGVK_NONE;	
};

std::string MGVK_Format(enum MGVirtualKeys mgvk) {
	static const char* lookup[MGVK_MAX_VIRTKEYS] = {
		"(none)",
		"Tab", "Enter", "Esc", "Space", "Ins", "~", "Del",
		"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
		"A", "B", "C", "D", "E", "F", "G", "H", "I", "J",
		"K", "L", "M", "N", "O", "P", "Q", "R", "S", "T",
		"U", "V", "W", "X", "Y", "Z",

		"F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9",
		"F10", "F11", "F12",

		// utf-8 for u,r,d,l arrows
		"\xE2\x86\x91", "\xE2\x86\x92", "\xE2\x86\x93", "\xE2\x86\x90", 
	};
	if (mgvk > 0 && mgvk < MGVK_MAX_VIRTKEYS) {
		return lookup[mgvk];
	}
	return "";
};

// Turn a key combination into a printable character, in a best-approximation of
//   what would happen if the keycombination was applied natively to a text input.
std::string MGKC_FormatCharacter(MGKeyCombination& kc) {
	// FIXME: all printable characters (hyphens, shift+numbers, ...)
	// It's never going to support custom IMEs though
	std::string ret;
	if (kc.virtkey >= MGVK_0 && kc.virtkey <= MGVK_Z) {
		ret = MGVK_Format(kc.virtkey);
		if (!kc.shift) {
			ret = tolower(ret[0]);
		}
	}
	return ret;
}

std::string MGKC_Format(MGKeyCombination& kc) {
	// Used for e.g. displaying hotkeys.
	std::string ret;
	if (kc.ctrl) ret += "Ctrl";	
	if (kc.shift) {
		if (ret.length()) ret += " + ";
		ret += "Shift";
	}
	if (kc.alt) {
		if (ret.length()) ret += " + ";
		ret += "Alt";
	}
	if (ret.length()) ret += " + ";
	ret += MGVK_Format(kc.virtkey);
	return ret;
};

std::string MGKC_FormatSmall(MGKeyCombination& kc) {
	// As per above, but with no spaces (ideally they would be mergable)
	// Used for e.g. displaying hotkeys in menu items.
	std::string ret;
	if (kc.ctrl) ret += "Ctrl";	
	if (kc.shift) {
		if (ret.length()) ret += "+";
		ret += "Shift";
	}
	if (kc.alt) {
		if (ret.length()) ret += "+";
		ret += "Alt";
	}
	if (ret.length()) ret += "+";
	if (kc.virtkey != MGVK_NONE) {
		ret += MGVK_Format(kc.virtkey);
	}
	return ret;
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Font manager
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGFontManager* MGFontManager::singleton = 0;

MGFontManager::MGFontManager() {
	hfMain = (HFONT)0;
	fontTwipSize = 0;
	bICCE = false;
	singleton = 0;
}

MGFontManager* MGFontManager::getInstance() {
	if (! singleton) singleton = new MGFontManager();
	return singleton;
}

const wchar_t* MGFontManager::getFontFace() {
	return fontFace.c_str();
}
int MGFontManager::getFontTwipHeight() {
	return fontTwipSize;
}
int MGFontManager::getFontPixelHeight() {
	return fontTwipSize / 20;
}

int MGFontManager::LogheightToTwips(int lfHeight) {
	HDC hDC = GetDC(GetDesktopWindow() );
	int PPI_Y = GetDeviceCaps(hDC, LOGPIXELSY);
	ReleaseDC( GetDesktopWindow(), hDC);
	return abs( (lfHeight*72*20) / PPI_Y);
}

void MGFontManager::SetControlFont(HWND hWnd) {
	if (hfMain==(HFONT)0) {

		// Retrieve system font
		NONCLIENTMETRICS ncmSystem = {0}; // Changes with NT6, so ensure WINVER is set..
		ncmSystem.cbSize = sizeof(ncmSystem);
		SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncmSystem), &ncmSystem, 0);
		hfMain = CreateFontIndirect(&ncmSystem.lfMessageFont);

		// Cache facename and height in twips
		fontFace.assign( ncmSystem.lfMessageFont.lfFaceName );
		fontTwipSize = LogheightToTwips(ncmSystem.lfMessageFont.lfHeight);
	}
	
	//Send new font
	SendMessage(hWnd, WM_SETFONT, (WPARAM)hfMain, 0); 
}

void MGFontManager::EnsureCommonControls() {
	if (!bICCE) {
		INITCOMMONCONTROLSEX icc = {0};
		icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
		icc.dwICC = ICC_WIN95_CLASSES | ICC_STANDARD_CLASSES | ICC_TAB_CLASSES;
		InitCommonControlsEx(&icc);
		bICCE = true;
	}
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGEventHandler
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGEventHandler::MGEventHandler(MGControl *master) {
	m_master = master;
	m_oldProc = 0;	
}

void MGEventHandler::subclassWnd(HWND hWnd) {
	if (m_oldProc) return; //already subclassed
	
	// Subclass and store old proc
	m_hWnd = hWnd;
	SetWindowLongPtr(m_hWnd, GWLP_USERDATA, (LONG)this);
	m_oldProc = SetWindowLongPtr(m_hWnd, GWLP_WNDPROC, (LONG)(&stProc));
}

LRESULT CALLBACK MGEventHandler::stProc( // Static
	HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam
) {
	if (uMsg == WM_NCCREATE) {
		SetWindowLongPtr(  // Store..
			hwnd, GWLP_USERDATA, (long)((LPCREATESTRUCT(lParam))->lpCreateParams)
		);
	}

	// ..Retrieve
	MGEventHandler* e = (MGEventHandler*) GetWindowLongPtr(hwnd, GWLP_USERDATA);
	
	if (e) return e->Proc(hwnd, uMsg, wParam, lParam);
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK MGEventHandler::Proc(
	HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam
) {
	MGEventData evdata = {0};
	evdata.hwnd   = hwnd;
	evdata.uMsg   = uMsg;
	evdata.wParam = wParam;
	evdata.lParam = lParam;

	// call control's own eventhandler function
	long ret = m_master->_parseEvent(&evdata);

	if (ret) {
		// The replacement procedure consumed the message.
		return 0; //ret;
	} else {
		// The replacement procedure didn't consume the message.
		if (m_oldProc) {
			// We overrode the proc;
			return CallWindowProc((WNDPROC)(m_oldProc), hwnd, uMsg, wParam, lParam);
		} else {
			// We didn't override the proc
			return DefWindowProc(hwnd, uMsg, wParam, lParam);
		}
	}
}

MGEventHandler::~MGEventHandler() {
	// TODO: If there was an original subclass proc, restore it
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGWindow
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGWindow::MGWindow(void* hIcon, const char* caption, MGWindow* modalParent) {
	
	OleInitialize(NULL);

	this->_controlType = MGCT_WINDOW;
	e = new MGEventHandler(this);
	//d = new MGControlData;
	w = new MGWindowData;
	d->hInstance = GetModuleHandle(0); //(HINSTANCE)hInstance;
	d->ID = 100; // Next control ID
	
	if (modalParent) {
		d->ParentHWnd = modalParent->d->hWnd;
		w->modalParent = modalParent;

		modalParent->setEnabled(false);
	} else {
		d->ParentHWnd = 0;
		w->modalParent = NULL;
	}
	
	w->controlIDs.clear();
	
	MGFontManager::getInstance()->EnsureCommonControls();

	// TODO: only register this class once per MGui
	WNDCLASS wcx = {0}; 
		wcx.style = CS_HREDRAW | CS_VREDRAW;
		wcx.lpfnWndProc = e->stProc;
		wcx.hInstance = d->hInstance;
		wcx.hIcon = (HICON)LoadIcon(d->hInstance, MAKEINTRESOURCE(hIcon)); //(HICON)hIcon
		wcx.hCursor = LoadCursor(NULL, IDC_ARROW);
		wcx.hbrBackground = (HBRUSH)(COLOR_WINDOW);
		wcx.lpszMenuName = NULL;
		wcx.lpszClassName = L"MGWindow";
	RegisterClass(&wcx);
	
	// Create window
	
	d->hWnd = CreateWindowEx(
		(d->ParentHWnd ? 0 : WS_EX_APPWINDOW),
		L"MGWindow", utf8to16(caption).c_str(), WS_OVERLAPPEDWINDOW & (~ WS_VISIBLE),
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, d->ParentHWnd,
		NULL, d->hInstance, (void*)e
	);
	w->bActive = d->hWnd?true:false; 
	if (! w->bActive ) return;

	MGFontManager::getInstance()->SetControlFont(d->hWnd);

	// Create status bar
	w->hStatus = CreateWindowEx(0, STATUSCLASSNAME, L"", WS_CHILD | WS_VISIBLE,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		d->hWnd, 0, d->hInstance, 0
	);
	SendMessage(w->hStatus, SB_SIMPLE, TRUE, 0);
	MGFontManager::getInstance()->SetControlFont(w->hStatus);
	this->showStatusBar(false);

	// Create menu
	w->hAccelTable = 0;
	w->hMenu = CreateMenu();
	w->childMenus.clear();
	SetMenu(d->hWnd, w->hMenu);

	// Show
	//ShowWindow(d->hWnd, false); // Don't show
	UpdateWindow(d->hWnd);
}

void MGWindow::close() {
	onClosing.invoke();
	//PostQuitMessage(0);
	DestroyWindow(d->hWnd);
	//EndDialog(d->hWnd, NULL);
	//SendMessage(d->hWnd, WM_CLOSE, 0, 0);
	
	w->bActive = false;
}

MGWindow::~MGWindow() {
	
	if (w->modalParent) {
		// Re-enable parent window before this window is destroyed
		w->modalParent->setEnabled(true);
	}

	SendMessage(d->hWnd, WM_DESTROY, 0, 0);
	w->bActive = false;
	if (w->hAccelTable) {
		DestroyAcceleratorTable(w->hAccelTable);
		w->hAccelTable = 0;
	}

	if (w->modalParent) {
		w->modalParent->focus();
	}

	delete w;
	//delete d;
	delete e;

	OleUninitialize();
}

BOOL MG_IsDialogMessage(HWND hWnd, LPMSG lpMsg) {
	if (lpMsg->message == WM_KEYDOWN && lpMsg->wParam == VK_RETURN) {
		return FALSE ; // Not handled
	} else {
		return ::IsDialogMessage(hWnd, lpMsg);
	}
};

bool MGWindow::doEvents() {
	MSG msg;
	bool evented = false;
	while (PeekMessage(&msg, NULL, 0, 0, 1) > 0) {
		evented = true;
		if (! TranslateAccelerator( d->hWnd, w->hAccelTable, &msg ) &&
			! MG_IsDialogMessage(d->hWnd, &msg)
		) {
			TranslateMessage(&msg); 
			DispatchMessage(&msg); 
		}
	}
	return evented;
}

void MGWindow::doEventLoop() {
	MSG msg;
	while (w->bActive && GetMessage(&msg, NULL, 0, 0) > 0) {
		if (! TranslateAccelerator( d->hWnd, w->hAccelTable, &msg ) &&
			! MG_IsDialogMessage(d->hWnd, &msg)
		) {
			TranslateMessage(&msg); 
			DispatchMessage(&msg); 
		}
	}
}

bool MGWindow::isActive() {
	return w->bActive;
}

void MGWindow::_prepareChild(MGControlData *pCtrlData, MGControl* pCtrl) {
	pCtrlData->hInstance = d->hInstance;
	pCtrlData->ParentHWnd = d->hWnd;
	pCtrlData->ID = d->ID ++;
	w->controlIDs[pCtrlData->ID] = pCtrl;
}

void MGWindow::_prepareChild(MGMenuItemData *pMenuData, MGMenuItem* pMenu) {
	pMenuData->ID = d->ID ++;
	pMenuData->parentHWnd = d->hWnd;
	w->controlIDs[pMenuData->ID] = pMenu;
}

long MGWindow::_parseEvent(MGEventData *ev) {

	static HBRUSH hbrChild = NULL;

	switch (ev->uMsg) {
		case WM_CLOSE: {
			close(); // invokes onClosing and calls DestroyWindow(hwnd)
		} break;

		case WM_COMMAND: {
			// wParam contains the ID, so look it up
			std::map<unsigned int, MGObject*>::iterator pObj =
				w->controlIDs.find(LOWORD(ev->wParam))
			;
			if (pObj != w->controlIDs.end()) {
				MGObject* ctrl = (*pObj).second;
				switch (ctrl->_controlType) {
					case MGCT_BUTTON: {
						((MGButton*)ctrl)->onClick.invoke();
					} break;
					case MGCT_CHECKBOX: {
						((MGCheckBox*)ctrl)->onCheck.invoke();
					} break;
					case MGCT_MENUITEM: {
						((MGMenuItem*)ctrl)->onClick.invoke();
					} break;
					case MGCT_LISTBOX: {
						if (HIWORD(ev->wParam) == LBN_SELCHANGE) {
							((MGListBox*)ctrl)->onChange.invoke();
						} else if (HIWORD(ev->wParam) == LBN_DBLCLK) {
							((MGListBox*)ctrl)->onDoubleClick.invoke();
						}
					} break;
					case MGCT_COMBOBOX: {
						if (HIWORD(ev->wParam) == CBN_SELCHANGE) {
							((MGComboBox*)ctrl)->onChange.invoke();
						}
					} break;
				}
			}
		} break;
		
		// Alt, shift, and control do not trigger the flag by themselves, only
		//   when used with other keys.
		// Even though WM_KEYDOWN is called for these keys
		case WM_KEYDOWN: {
			enum MGVirtualKeys vk = MGNativeToVK(ev->wParam);
			if (vk != MGVK_NONE) {
				
				// We could cache these as flags when WM_KEYDOWN happens for
				//  them.. or just look them up on the spot
				SHORT a = GetKeyState(VK_MENU),
					c = GetKeyState(VK_CONTROL),
					s = GetKeyState(VK_SHIFT);

				onKey.data = MGKC(vk, !!(a & 0x80), !!(c & 0x80), !!(s & 0x80));
				onKey.invoke();

				if (onKey.wasEventDefaultPrevented()) {
					return true;
				}
			}
		} break;

		case WM_SIZE: {
			onResize.invoke();
			SendMessage(w->hStatus, WM_SIZE, 0, 0);
		} break;

		case WM_MOVE: {
			onMove.invoke();
		} break;

		case WM_TIMER: {
			std::map<unsigned int, MGObject*>::iterator pObj =
				w->controlIDs.find(LOWORD(ev->wParam))
			;
			if (pObj != w->controlIDs.end()) {
				MGObject* ctrl = (*pObj).second;
				if (ctrl->_controlType == MGCT_TIMER) {
					((MGTimer*)ctrl)->onTick.invoke();
				}
			}
		} break;

		case WM_HOTKEY: {
			std::map<unsigned int, MGObject*>::iterator pObj =
				w->controlIDs.find(LOWORD(ev->wParam))
			;
			if (pObj != w->controlIDs.end()) {
				MGObject* ctrl = (*pObj).second;
				if (ctrl->_controlType == MGCT_HOTKEY) {
					((MGHotkey*)ctrl)->onPress.invoke();
					if ( ((MGHotkey*)ctrl)->onPress.wasEventDefaultPrevented()) {
						return true;
					}
				}
			}
		} break;
		
		case WM_MEASUREITEM:
		case WM_DRAWITEM:
		{ // These are listbox calls that get sent to the parent. Forward them on
			std::map<unsigned int, MGObject*>::iterator pObj =
				w->controlIDs.find(LOWORD(ev->wParam))
			;
			if (pObj != w->controlIDs.end()) {
				MGObject* ctrl = (*pObj).second;
				if (ctrl->_controlType == MGCT_LISTBOX) {
					return ((MGListBox*)ctrl)->_parseEvent(ev);
				}
			}
		} break;

		case WM_USER: {
			// WPARAM: event id, LPARAM: void* data
			std::map<unsigned int, MGAction<long>*>::iterator pCB = 
				w->events.find( ev->wParam);

			if (pCB != w->events.end()) {
				pCB->second->data = ev->lParam;
				pCB->second->invoke();
				if (pCB->second->wasEventDefaultPrevented()) {
					return true;
				}
			}
		} break;
		
		case WM_CTLCOLORSTATIC: {
			// Fix background colors for labels on Windows 8
			HDC hdcStatic = (HDC)ev->wParam;
			SetBkColor(hdcStatic, GetSysColor(COLOR_MENU));
			if (hbrChild == NULL) {
				hbrChild = CreateSolidBrush(GetSysColor(COLOR_MENU));
			}
			return (LONG_PTR)hbrChild;
		} break;
		
		case WM_DROPFILES: {
			HDROP hDrop = (HDROP)ev->wParam;

#define DRAGQUERYFILE_GET_ALL_FILES (0xFFFFFFFF)

			UINT numFiles = DragQueryFile(hDrop, DRAGQUERYFILE_GET_ALL_FILES, NULL, NULL);
			std::vector<std::string> filenames;
			filenames.reserve(numFiles);

			for (size_t i = 0; i != numFiles; ++i) {
				wchar_t tBuff[MAX_PATH+1] = {0};
				UINT charct = DragQueryFile(hDrop, i, tBuff, MAX_PATH);
				filenames.push_back( utf16to8(tBuff) );
			}

			onDropFiles.data.swap(filenames);
			onDropFiles.invoke();
			onDropFiles.data.clear();

			DragFinish(hDrop);
		} break;

		default: {
		} break;
	}
	return false; // not handled
}

void MGWindow::addChildMenu(MGMenuItem* menuItem) {
	w->childMenus.push_back(menuItem);
}

void MGWindow::updateMenu() {
	// Clear all existing items
	if (w->hMenu) DestroyMenu(w->hMenu);
	w->hMenu = CreateMenu();

	// Destroy current accelerator table
	if (w->hAccelTable) {
		DestroyAcceleratorTable(w->hAccelTable);
		w->hAccelTable = 0;
	}
	w->childMenuAccels.clear();

	// Add all child elements, and all their children, etc
	MGMenuItemRelation relation;
	relation.parent = w->hMenu;
	relation.menuAccels = &w->childMenuAccels;
	for (auto i = w->childMenus.begin(), e = w->childMenus.end(); i != e; i++) {
		(*i)->_updateSelf(&relation);
	}

	// Recreate accelerator table
	if ( w->childMenuAccels.size() ) {
		w->hAccelTable = CreateAcceleratorTable(
			&w->childMenuAccels[0], w->childMenuAccels.size()
		);
	}

	// Re-associate to window
	SetMenu(d->hWnd, w->hMenu);
}

void MGWindow::focus() {
	// Un-minimise if minimised
	::ShowWindow(d->hWnd, SW_RESTORE);

	// Float to top
	BringWindowToTop(d->hWnd);
	SetForegroundWindow(d->hWnd);
}

void MGWindow::setResizable(bool resizable) {
	long style = GetWindowLongPtr(d->hWnd, GWL_STYLE);
	if (resizable) {
		style |= WS_THICKFRAME | WS_MAXIMIZEBOX;
	} else {
		style &= ~ WS_THICKFRAME;
		style &= ~ WS_MAXIMIZEBOX;
	}
	SetWindowLongPtr(d->hWnd, GWL_STYLE, style);
}

void MGWindow::showStatusBar(bool show) {
	if (show) {
		DestroyWindow(w->hStatus);
			w->hStatus = CreateWindowEx(0, STATUSCLASSNAME, L"", WS_CHILD | WS_VISIBLE,
			CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
			d->hWnd, 0, d->hInstance, 0
		);
		SendMessage(w->hStatus, SB_SIMPLE, TRUE, 0);
		MGFontManager::getInstance()->SetControlFont(w->hStatus);
		//this->showStatusBar(false);
	}

	ShowWindow(w->hStatus, show ? SW_SHOW : SW_HIDE);
	w->bStatusBarVisible = show;
}

void MGWindow::setStatusBar(const char* text) {
	SendMessageW(w->hStatus, SB_SETTEXT, SB_SIMPLEID, (LPARAM)(utf8to16(text).c_str()));	
}

int MGWindow::getInnerHeight() {
	// We may have to take into account statusbar height
	RECT rct; GetClientRect( d->hWnd, &rct );
	
	if (w->bStatusBarVisible) {
		RECT rctStatus;
		GetClientRect( w->hStatus, &rctStatus);
		return rct.bottom - rct.top - rctStatus.bottom + rctStatus.top;
	} else {
		return rct.bottom - rct.top;
	}
}

void MGWindow::setCaption(const char* caption) {
	SetWindowText(d->hWnd, utf8to16(caption).c_str());
}

void MGWindow::registerMarshall(unsigned int event_id, MGAction<long>* doCall) {
	w->events[event_id] = doCall;
}

bool MGWindow::doMarshalled(unsigned int event_id, long data) {
	if (w->events.find(event_id) != w->events.end()) {
		PostMessage(d->hWnd, WM_USER, event_id, data);
		return true;
	} else {
		return false;
	}
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGWindow: Dialog calls
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


void MGWindow::dlgAlert(const char* message, const char* caption) {
	MessageBox(d->hWnd, utf8to16(message).c_str(), utf8to16(caption).c_str(), MB_OK);
}

void MGWindow::dlgAlert(const char* message) {
	wchar_t captionBuffer[256] = {0};
	GetWindowText(d->hWnd, captionBuffer, 255);
	MessageBox(d->hWnd, utf8to16(message).c_str(), captionBuffer, MB_OK);
}

bool MGWindow::dlgConfirm(const char* prompt) {
	wchar_t captionBuffer[256] = {0};
	GetWindowText(d->hWnd, captionBuffer, 255);
	return MessageBox(d->hWnd, utf8to16(prompt).c_str(), captionBuffer, MB_YESNO) == IDYES;
}
/*
bool MGWindow::dlgFileOpen(const char* title, const char* filter, std::string &result) {
	wchar_t retbuff[512] = {0};
	std::wstring wtitle = utf8to16(title), wfilter;

	// Build key\0filter\0key\0filter\0\0 from key|filter|key|filter

	std::vector<std::string> filter_parts = explode("|", std::string(filter));
	for (int i = 0, e = 2 * (filter_parts.size() / 2); i < e; i++) {
		wfilter.append(utf8to16(filter_parts[i]));
		wfilter.append(L"\0", 1);
	}
	wfilter.append(L"\0\0", 2);

	OPENFILENAME ofn = {0};
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = d->hWnd;
	ofn.lpstrFilter = wfilter.c_str(); 
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = retbuff;
	ofn.nMaxFile = 512 - 1;
	ofn.lpstrTitle = wtitle.c_str();
	ofn.Flags = OFN_DONTADDTORECENT | OFN_EXPLORER | OFN_FORCESHOWHIDDEN;

	if (! GetOpenFileName(&ofn)) return false; // cancelled

	result = utf16to8(ofn.lpstrFile);
	return true;
}
*/
bool MGWindow::dlgFolderSelect(const char* title, std::string &result) {
	// TODO: MSDN says we need to intialise COM first..?
	// Calling this once seems to allocate a lot of memory that isn't freed.
	// Although, at least calling it multiple times is stable.

	CoInitialize(NULL);

	bool retval = false;
	std::wstring wTitle = utf8to16(title);
	do {
		wchar_t resultBuffer[MAX_PATH] = {0};
	
		BROWSEINFO binf = {0};
		binf.hwndOwner = d->hWnd;
		binf.lpszTitle = wTitle.c_str();
		binf.ulFlags = BIF_NEWDIALOGSTYLE | BIF_NONEWFOLDERBUTTON;

		ITEMIDLIST* items = SHBrowseForFolder(&binf);
		if (! items) break; // User cancelled
	
		if (! SHGetPathFromIDList(items, resultBuffer)) break;
		
		IMalloc *imalloc = 0;
		if(SUCCEEDED(SHGetMalloc(&imalloc))) {
			imalloc->Free(items);
			imalloc->Release();
		}

		result.assign(utf16to8(resultBuffer));
		retval = result.length() > 0;
	} while (false);
	
	CoUninitialize();

	return retval;
}

// Folder select with custom checkbox..

struct FolderSelectExInfo {
	const char* optname;
	bool optresult;
	HWND hDialog, hCheckbox;
};
#define IDD_DLG_CHECKBOX 150
static WNDPROC gOriginalDlgProc = NULL;
static FolderSelectExInfo* gFSXI = NULL;

LRESULT CALLBACK dlgFolderNewProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch(uMsg) {
		case WM_SIZE: {
			// move dialog
			unsigned short width = LOWORD(lParam), height = HIWORD(lParam);

			SetWindowPos(
				gFSXI->hCheckbox, 0, 10, height - 35, width - 185, 25,
				SWP_NOZORDER
			);
		} break;

		case WM_COMMAND: {
			if ((HWND)lParam == gFSXI->hCheckbox) {
				if (SendMessage(gFSXI->hCheckbox, BM_GETCHECK, 0, 0) == BST_CHECKED) {
					gFSXI->optresult = true;
				} else {
					gFSXI->optresult = false;
				}
			}
		} break;
	}
	return CallWindowProc(gOriginalDlgProc, hwnd, uMsg, wParam, lParam);
}

static int CALLBACK dlgFolderExProc(HWND hWnd, UINT uMsg, LPARAM lParam, LPARAM lpData) {
	//FolderSelectExInfo *fsxi = (FolderSelectExInfo*)lpData;

	switch(uMsg) {
		case BFFM_INITIALIZED: {

			gFSXI->hDialog = hWnd;

			// Create checkbox

			std::wstring loptname = utf8to16( gFSXI->optname );

			gFSXI->hCheckbox = CreateWindow(
				L"BUTTON", loptname.c_str(), 
				WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_TEXT | BS_VCENTER | BS_AUTOCHECKBOX,
				10, 10, 300, 25, hWnd, 0, 0, NULL
			);
			MGFontManager::getInstance()->SetControlFont(gFSXI->hCheckbox);
			
			// Subclass this dialog
			gOriginalDlgProc = (WNDPROC) SetWindowLong(
				hWnd, DWL_DLGPROC, (LONG)dlgFolderNewProc
			);

		} break;
	}
	return 1;
}

bool MGWindow::dlgFolderSelectEx(
	const char* title,
	const char* optname,
	std::string &result,
	bool &optresult
){
	gFSXI = new FolderSelectExInfo;
	gFSXI->optname = optname;
	gFSXI->optresult = false; // we made the checkbox default-unchecked

	bool retval = false;
	std::wstring wTitle = utf8to16(title);
	do {
		wchar_t resultBuffer[MAX_PATH] = {0};
	
		BROWSEINFO binf = {0};
		binf.hwndOwner = d->hWnd;
		binf.lpszTitle = wTitle.c_str();
		binf.ulFlags = BIF_NEWDIALOGSTYLE | BIF_NONEWFOLDERBUTTON;
		binf.lpfn = dlgFolderExProc;
		// binf.lParam = (LPARAM) &gFSXI; // doesn't... seem to work

		ITEMIDLIST* items = SHBrowseForFolder(&binf);
		if (! items) break; // User cancelled
	
		if (! SHGetPathFromIDList(items, resultBuffer)) break;
		
		IMalloc *imalloc = 0;
		if(SUCCEEDED(SHGetMalloc(&imalloc))) {
			imalloc->Free(items);
			imalloc->Release();
		}

		result.assign(utf16to8(resultBuffer));
		retval = result.length() > 0;
	} while (false);
	
	optresult = gFSXI->optresult;
	delete gFSXI; gFSXI=NULL;

	return retval;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGMenuItem
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGMenuItem::MGMenuItem(MGWindow* parent, const char* text) {
	_initialise(parent, text);
	parent->addChildMenu(this);
}

MGMenuItem::MGMenuItem(MGMenuItem* parent, const char* text) {
	_initialise(parent->m->parentWnd, text);
	parent->addChildMenu(this);
}

MGMenuItem::MGMenuItem(MGWindow* parent, const char* text, MGCallback onClick) {
	_initialise(parent, text);
	this->onClick.assign(parent, onClick);
	parent->addChildMenu(this);
}

MGMenuItem::MGMenuItem(	MGMenuItem* parent, const char* text, MGCallback onClick) {
	_initialise(parent->m->parentWnd, text);
	this->onClick.assign(parent->m->parentWnd, onClick);
	parent->addChildMenu(this);
}

MGMenuItem::MGMenuItem(MGMenuItem* parent, const char* text, MGCallback onClick, MGKeyCombination shortcut) {
	_initialise(parent->m->parentWnd, text);
	this->onClick.assign(parent->m->parentWnd, onClick);
	parent->addChildMenu(this);
	setShortcut(shortcut);
}

void MGMenuItem::_initialise(MGWindow* parent, const char* text) {
	this->_controlType = MGCT_MENUITEM;
	m = new MGMenuItemData;
	parent->_prepareChild(m, this);
	m->caption.assign(utf8to16(text).c_str());
	m->hMenu = 0; // For child menus
	m->shortcut.virtkey = MGVK_NONE; // invalid/notset
	m->parentWnd = parent;
}

MGMenuItem::~MGMenuItem() {
	if (m) {
		if (m->hMenu) {
			DestroyMenu(m->hMenu);
			m->hMenu = 0;
		}
		delete m;
		m = 0;
	}
}

std::string MGMenuItem::getText() {
	return utf16to8(m->caption.c_str());
}

void MGMenuItem::setText(std::string& sz) {
	// updateMenu() required for changes to take affect
	m->caption.assign(utf8to16(sz.c_str()));
}

MGMenuItem* MGMenuItem::setShortcut(MGKeyCombination kc) {
	
	// Update menu item text
	// (FIXME: what if this is called twice...? Check if a keycombination is
	//   already set for this menu item and strip back from the last L"\t")

	setText(
		getText() + "\t" + MGKC_FormatSmall(kc)
	);

	// Assign shortcut
	m->shortcut = kc;
	return this; // Chainable!
}

MGMenuItem* MGMenuItem::addChildMenu(MGMenuItem* menuItem) {
	m->children.push_back(menuItem);
	return this;
}

MGMenuItem* MGMenuItem::addChildMenuSeparator() {
	m->children.push_back( (MGMenuItem*) 0 );	
	return this;
}

MGMenuItem* MGMenuItem::setChecked(bool isChecked) {
	// FIXME: doesn't work if updateMenu() hasn't already been called...
	CheckMenuItem(GetMenu(m->parentHWnd), m->ID, MF_BYCOMMAND | (isChecked ? MF_CHECKED : MF_UNCHECKED));
	DrawMenuBar(m->parentHWnd);
	return this;
}

MGMenuItem* MGMenuItem::setEnabled(bool isEnabled) {
	EnableMenuItem(GetMenu(m->parentHWnd), m->ID, MF_BYCOMMAND | (isEnabled ? MF_ENABLED : MF_GRAYED));
	DrawMenuBar(m->parentHWnd);
	return this;
}

void MGMenuItem::_updateSelf(MGMenuItemRelation *relation) {
	// Add shortcut accelerator (if necessary)
	if (m->shortcut.virtkey) {
		ACCEL myAccel = {
			FVIRTKEY | 
			(m->shortcut.alt ? FALT : 0) |
			(m->shortcut.ctrl ? FCONTROL : 0) |
			(m->shortcut.shift ? FSHIFT : 0),
			MGVKToNative( m->shortcut.virtkey ),
			m->ID
		};
		relation->menuAccels->push_back(myAccel);
	}

	DWORD dFlags = MF_ENABLED | MF_UNCHECKED | MF_STRING;
	if (m->children.size() > 0) {
		if (! m->hMenu) m->hMenu = CreatePopupMenu();
		// create all children
		for (auto i = m->children.begin(), e = m->children.end(); i != e; i++) {
			if ( (*i) == 0 ) { // it's a separator
				AppendMenu(m->hMenu, dFlags | MF_SEPARATOR, 0, 0);
			} else { // a real child
				MGMenuItemRelation childRelation;
				childRelation.parent = m->hMenu;
				childRelation.menuAccels = relation->menuAccels;
				(*i)->_updateSelf(&childRelation);
			}
		}
		// create as foldout menu
		AppendMenu(
			relation->parent, dFlags | MF_POPUP, (UINT_PTR)m->hMenu, m->caption.c_str()
		);
	} else {
		// create as link
		AppendMenu(relation->parent, dFlags, m->ID, m->caption.c_str() );
	}
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGControl
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGControl::MGControl() {
	d = new MGControlData;
}

MGControl::~MGControl() {
	if (d) delete d;
}

void MGControl::resize(int x, int y, int width, int height) {
	SetWindowPos(d->hWnd, 0, x, y, width, height, SWP_NOZORDER);
}

void MGControl::resize(int width, int height) {
	SetWindowPos(d->hWnd, 0, 0, 0, width, height, SWP_NOZORDER | SWP_NOMOVE);
}

void MGControl::move(int x, int y) {
	SetWindowPos(d->hWnd, 0, x, y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
}

#define SNAP_THRESHOLD (10)
void MGControl::snapToScreenEdge() {
	RECT primaryDisplay = {0};
	//GetWindowRect(GetDesktopWindow(), &primaryDisplay); // occludes taskbar
	SystemParametersInfo(SPI_GETWORKAREA, 0, &primaryDisplay, 0);

	// FIXME: The current algorithm causes some too-far-overlap when fullscreening
	//  the window. It would be better to have a general-purpose solution for
	//  this but in the mean time, disable doing this if the window is fullscreened

	if (getInnerWidth() == primaryDisplay.right - primaryDisplay.left) {
		// Assume that means we're fullscreened. The inner height isn't going
		//  to be the same because of the top taskbar, and the outer dimensions
		//  aren't going to be the same because they stretch so that the 
		//  client area covers the screen
		// Really should check to see if there's a getWindowXXXX for this, or
		//  cache WM_SIZE wParam==SIZE_MAXIMISED
		return;
	}

	int x = getX(),
		y = getY(),
		oh = getHeight(),
		ow = getWidth();

	if (abs(x - primaryDisplay.left) < SNAP_THRESHOLD) {
		move(0, y);
	} else if (abs((x + ow) - primaryDisplay.right) < SNAP_THRESHOLD) {
		move(primaryDisplay.right - ow, y);
	}

	// Now don't use x as it may have changed
	if (abs(y - primaryDisplay.top) < SNAP_THRESHOLD) {
		move(getX(), 0);
	} else if (abs((y + oh) - primaryDisplay.bottom) < SNAP_THRESHOLD) {
		move(getX(), primaryDisplay.bottom - oh);
	}
}

void MGControl::refresh() {
	InvalidateRect(d->hWnd, NULL, false);
}

void MGControl::focus() {
	if (d->hWnd) SetFocus(d->hWnd);
}

void MGControl::setEnabled(bool bEnabled) {
	EnableWindow(d->hWnd, (bEnabled ? TRUE : FALSE ));
}

void MGControl::setVisible(bool bVisible) {
	/*if (_controlType == MGCT_WINDOW) {
		ShowWindow(d->hWnd, (bVisible ? SW_SHOWNORMAL : SW_HIDE) );
	} else {*/
		ShowWindow(d->hWnd, (bVisible ? SW_SHOW : SW_HIDE) );
	//}
}

int MGControl::getX() {
	RECT rct; GetWindowRect( d->hWnd, &rct );
	return rct.left;
}

int MGControl::getY() {
	RECT rct; GetWindowRect( d->hWnd, &rct );
	return rct.top;
}

int MGControl::getWidth() {
	RECT rct; GetWindowRect( d->hWnd, &rct );
	return rct.right - rct.left;
}

int MGControl::getHeight() {
	RECT rct; GetWindowRect( d->hWnd, &rct );
	return rct.bottom - rct.top;
}

int MGControl::getInnerWidth() {
	RECT rct; GetClientRect( d->hWnd, &rct );
	return rct.right - rct.left;
}

int MGControl::getInnerHeight() {
	RECT rct; GetClientRect( d->hWnd, &rct );
	return rct.bottom - rct.top;
}

long MGControl::_parseEvent(MGEventData* ev) {
	return false; // Default; no action
}

void MGControl::removeStyles() {
	SetWindowTheme(d->hWnd, L"", L"");
}

void MGControl::removeBorder() {
	long style = GetWindowLong(d->hWnd, GWL_EXSTYLE);
	SetWindowLong(d->hWnd, GWL_EXSTYLE, style & ~WS_EX_STATICEDGE);
}

void MGControl::refreshHasDropFilesHandler() {
	if (onDropFiles.isAssigned()) {
		DragAcceptFiles(d->hWnd, TRUE);
	} else {
		DragAcceptFiles(d->hWnd, FALSE);
	}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGEdit
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGEdit::MGEdit(MGWindow* parent, bool bMultiline, bool isPassword) {
	this->_controlType = MGCT_EDIT;
	//d = new MGControlData;
	parent->_prepareChild(d, this);
	
	d->hWnd = CreateWindowEx(WS_EX_CLIENTEDGE, L"EDIT", 0,
		WS_VISIBLE | WS_CHILD | WS_TABSTOP | ES_AUTOVSCROLL |
		ES_AUTOHSCROLL | (bMultiline? ES_MULTILINE:0) | 
		(isPassword ? ES_PASSWORD : 0)
		,
		0, 0, 320, 240, d->ParentHWnd, (HMENU) d->ID, d->hInstance, NULL
	);
	if (!d->hWnd) return;

	e = new MGEventHandler(this);
	e->subclassWnd(d->hWnd);
	
	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

MGEdit::~MGEdit() {
	if (e) delete e;
}

long MGEdit::_parseEvent(MGEventData *ev) {	
	switch (ev->uMsg) {
		case WM_KEYUP: {
			MGVirtualKeys vk = MGNativeToVK(ev->wParam);
			
			onKey.data = MGKC(vk, false, false, false); // FIXME
			onKey.invoke();
			if (onKey.wasEventDefaultPrevented()) {
				return true;
			}
		} break;
	}
	return false; // unhandled, pass on
}

void MGEdit::textSet(const char* szText) {
	SetDlgItemText(d->ParentHWnd, d->ID, utf8to16(szText).c_str());
}

void MGEdit::textAppend(const char* szText) {
	// There might be a better way of doing this on platforms other than win32
	textSet( (textGet() + szText).c_str() );
	selEnd();
}

std::string MGEdit::textGet() {
	int textlen = textLength();
	if (! textlen) return std::string();
	textlen++;
	std::string ret;

	wchar_t* buff = new wchar_t[textlen + 1]; {

		GetDlgItemText(d->ParentHWnd, d->ID, buff, textlen);
		ret = utf16to8(buff);

	} delete[] buff;

	return ret;
}

int MGEdit::textLength() {
	return GetWindowTextLength(d->hWnd);
}

void MGEdit::selRange(int start, int end) {
	SendMessageW(d->hWnd, EM_SETSEL, start, end);
}

void MGEdit::selEnd() {
	// This trick is necessary (and correct)
	SendMessageW(d->hWnd, EM_SETSEL, 0,  -1);
	SendMessageW(d->hWnd, EM_SETSEL, -1, -1);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGButton
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGButton::MGButton(MGWindow *parent, const char *szCaption) {
	this->_controlType = MGCT_BUTTON;
	//d = new MGControlData;
	parent->_prepareChild(d, this);
	
	d->hWnd = CreateWindow(
		L"BUTTON", utf8to16(szCaption).c_str(), 
		WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_TEXT | BS_VCENTER,
		0, 0, 1, 1, d->ParentHWnd, (HMENU) d->ID, d->hInstance, NULL
	);

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGCheckBox
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGCheckBox::MGCheckBox(MGWindow *parent, const char *szCaption) {
	this->_controlType = MGCT_CHECKBOX;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(
		L"BUTTON", utf8to16(szCaption).c_str(), 
		WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_TEXT | BS_VCENTER | BS_AUTOCHECKBOX,
		0, 0, 1, 1, d->ParentHWnd, (HMENU) d->ID, d->hInstance, NULL
	);

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

bool MGCheckBox::isChecked() {
	return (IsDlgButtonChecked(d->ParentHWnd, d->ID) == BST_CHECKED);
};

void MGCheckBox::setChecked(bool checked) {
	CheckDlgButton(d->ParentHWnd, d->ID, checked ? BST_CHECKED : BST_UNCHECKED);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGListBox
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGListBox::MGListBox(MGWindow *parent, bool bAutoSort, bool bAllowColors, int lineheight, bool bAllowMultiSelect) :
	z(new MGListBoxData())
{
	this->_controlType = MGCT_LISTBOX;
	//d = new MGControlData;
	parent->_prepareChild(d, this);
	
	z->line_count = lineheight;

	d->hWnd = CreateWindowEx(
		WS_EX_STATICEDGE, L"LISTBOX", 0, 
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_TABSTOP | 
		LBS_NOINTEGRALHEIGHT | LBS_NOTIFY | LBS_USETABSTOPS | LBS_HASSTRINGS | LBS_WANTKEYBOARDINPUT |
		(bAutoSort ? LBS_SORT : 0) |
		(bAllowColors ? LBS_OWNERDRAWFIXED : 0) |
		(bAllowMultiSelect ? LBS_EXTENDEDSEL /*LBS_MULTIPLESEL*/ : 0)
		,
		0, 0, 1, 1, d->ParentHWnd, (HMENU) d->ID, d->hInstance, NULL
	);
	if (!d->hWnd) return;

	MGFontManager::getInstance()->SetControlFont(d->hWnd);

	if (bAllowColors) {
		z->default_color = COLOR_WINDOWTEXT;
	}
	z->is_multiselect = bAllowMultiSelect;
	z->allow_type_to_select = true;
	
	e = new MGEventHandler(this);
	e->subclassWnd(d->hWnd);
}

MGListBox::~MGListBox() {
	if (z) delete z;
	if (e) delete e;
}

unsigned int MGListBox::getSelected() {
	// Can return NOTHING_SELECTED = -1 as unsigned int 0xFFFFFFFF
	// Zero is a valid selection index, so comparisons to NOTHING_SELECTED need
	//   to be exact
	return (unsigned int) SendMessage(d->hWnd, LB_GETCURSEL, 0, 0);
}

unsigned int MGListBox::getSelectedCount() {
	if (z->is_multiselect) {
		return (unsigned int) SendMessage(d->hWnd, LB_GETSELCOUNT, 0, 0);
	} else {
		return (getSelected() == NOTHING_SELECTED) ? 0 : 1;
	}
}

void MGListBox::getAllSelected(std::vector<unsigned int>& idxList) {
	idxList.clear();
	if (z->is_multiselect) {
		unsigned int count = (unsigned int)SendMessage(d->hWnd, LB_GETSELCOUNT, 0, 0);

		// alloc
		unsigned int* ubuff = new unsigned int[count];

		// request
		unsigned int retCt = (unsigned int)SendMessage(d->hWnd, LB_GETSELITEMS, count, (LPARAM)ubuff);

		// copy (is there a std::vector constructor for this? std::copy maybe?)
		idxList.reserve(retCt);
		for (unsigned int i = 0, e = retCt; i != e; ++i) {
			idxList.push_back( ubuff[i] );
		}

		// free
		delete[] ubuff;

	} else {
		unsigned int curSel = getSelected();
		if (curSel != NOTHING_SELECTED) {
			idxList.push_back( curSel );
		}
	}
}

void MGListBox::setSelected(unsigned int idx) {
	if (z->is_multiselect) {
		// Clear all other selections
		SendMessage(d->hWnd, LB_SETSEL, FALSE, -1);
		SendMessage(d->hWnd, LB_SETSEL, TRUE, idx);
	} else {
		SendMessage(d->hWnd, LB_SETCURSEL, idx, 0);
	}
}

void MGListBox::multi_setSelected(unsigned int idx, bool isSelected) {
	if (z->is_multiselect) {
		SendMessage(d->hWnd, LB_SETSEL, isSelected ? TRUE : FALSE, idx);
	} else if (isSelected) {
		setSelected(idx);
	} else {
		setSelected(-1); // clear selections
	}
}

/*void MGListBox::setMultiline(int lines) {
	z->line_count = lines;
}*/

unsigned int MGListBox::addItem(const char* szItem) {
	return addItem_UTF16LE( utf8to16(szItem).c_str() );
}

unsigned int MGListBox::addItem_UTF16LE(const wchar_t* szItem) {
	return SendMessage(d->hWnd, LB_ADDSTRING, 0, (LPARAM)szItem);
}

unsigned int MGListBox::addItem(const char* szItem, unsigned int pos) {
	return addItem_UTF16LE( utf8to16(szItem).c_str(), pos );
}

unsigned int MGListBox::addItem_UTF16LE(const wchar_t* szItem, unsigned int pos) {
	unsigned int ret = SendMessage(d->hWnd, LB_INSERTSTRING, (WPARAM)pos, (LPARAM)szItem);

	if (z->colors.size()) {
		// Need to totally re-key the colors object.
		std::map<unsigned int, unsigned int> newcolors;

		for (std::map<unsigned int, unsigned int>::iterator i = z->colors.begin(), e = z->colors.end(); i != e; ++i) {
			if (i->first >= pos) {
				// above/equal - add
				newcolors[i->first + 1] = i->second;
			} else {
				// below - no change
				newcolors[i->first] = i->second;
			}
		}

		z->colors.swap(newcolors);
	}

	return ret;
}

unsigned int MGListBox::findItem(const char* szItem) {
	return (unsigned int) SendMessage(d->hWnd, LB_FINDSTRING, -1, (LPARAM)(utf8to16(szItem).c_str())); // -1 on error
}

unsigned int MGListBox::getItemCount() {
	return (unsigned int) SendMessage(d->hWnd, LB_GETCOUNT, 0, 0);
}

// WARNING: caller is responsible for free'ing any listitem data!
void MGListBox::clear() {
	SendMessage(d->hWnd, LB_RESETCONTENT, 0, 0);
	z->colors.clear();
}

// WARNING: caller is responsible for free'ing any listitem data!
void MGListBox::removeItem(unsigned int idx) {
	SendMessage(d->hWnd, LB_DELETESTRING, idx, 0);

	if (z->colors.size()) {
		// Need to totally re-key the colors object.
		std::map<unsigned int, unsigned int> newcolors;

		// Linear...
		for (std::map<unsigned int, unsigned int>::iterator i = z->colors.begin(), e = z->colors.end(); i != e; ++i) {
			if (i->first > idx) {
				// above - subtract
				newcolors[i->first - 1] = i->second;
			} else if (i->first == idx) {
				// removed
			} else {
				// below - no change
				newcolors[i->first] = i->second;
			}
		}

		z->colors.swap(newcolors);
	}
}

std::string MGListBox::getItemText(unsigned int idx) {
	std::string ret;

	int len = (int)SendMessage(d->hWnd, LB_GETTEXTLEN, idx, 0); // charcount, or -1 error
	if (len == -1) {
		ret.assign("");
	} else {
		wchar_t *buff = new wchar_t[len+1];
		SendMessage(d->hWnd, LB_GETTEXT, idx, (LPARAM)buff);
		ret.assign(utf16to8(buff));
		delete[] buff;
	}
	return ret;
}

void MGListBox::setItemText(unsigned int idx, const char* szItem) {
	// Bizzarely, there's no LB_SETTEXT / LB_SETITEMTEXT
	// At least doing things in this order preserves our colourmap

	// Don't use addItem/removeItem as each will traverse the colormap
	//  unnecessarily!

	void* itemdata = this->getItemData(idx);
	//this->removeItem(idx);
	SendMessage(d->hWnd, LB_DELETESTRING, idx, 0);
	//int real_idx = this->addItem(szItem, idx);
	unsigned int real_idx = SendMessage(
		d->hWnd, LB_INSERTSTRING, (WPARAM)idx, (LPARAM)(utf8to16(szItem).c_str())
	);

	this->setItemData(real_idx, itemdata);
}

void MGListBox::setItemData(unsigned int idx, void* data) {
	SendMessageW(d->hWnd, LB_SETITEMDATA, idx, (WPARAM)data);
}

void* MGListBox::getItemData(unsigned int idx) {
	return (void*) SendMessageW(d->hWnd, LB_GETITEMDATA, (WPARAM)idx, 0);
}

// Return an item to the default color. More efficient than setting the color
//  for this specific index - the implementation is more efficient at arbitrary 
//  placed inserts/deletes if fewer items have custom colors..
void MGListBox::setItemColor(unsigned int idx) {
	//z->colors[idx] = z->default_color;

	std::map<unsigned int, unsigned int>::iterator itCol = z->colors.find(idx);
	if (itCol != z->colors.end()) {
		z->colors.erase(idx);
	}
}

void MGListBox::setItemColor(unsigned int idx, unsigned char r, unsigned char b, unsigned char g) {
	z->colors[idx] = RGB(r,g,b);
}

void MGListBox::setAllowTypeToSelect(bool allowTypeToSelect) {
	z->allow_type_to_select = allowTypeToSelect;
}

long MGListBox::_parseEvent(MGEventData *ev) {
	switch(ev->uMsg) {
		case WM_DRAWITEM: {
	
			DRAWITEMSTRUCT* dis = (DRAWITEMSTRUCT*)ev->lParam;
			std::wstring text = utf8to16(getItemText(dis->itemID).c_str());

			HBRUSH brBack = 0;
			if (dis->itemState & ODS_SELECTED /*ODS_FOCUS*/) {

				// Item has focus, so highlight it using default focus colours
				brBack = GetSysColorBrush(COLOR_HIGHLIGHT);
				FillRect(dis->hDC, &(dis->rcItem), brBack);
				SetBkColor(dis->hDC, GetSysColor(COLOR_HIGHLIGHT));
				SetTextColor(dis->hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
				//DrawFocusRect(dis->hDC, &(dis->rcItem));
				// DrawFocusRect is also needed for the "correct" way of drawing the highlight,
				// but Billy/VCL doesn't do it, and i don't really like the way it looks anyway
				// If we do DrawFocusRect, then maybe increase the padding by one

			} else {

				// Item doesn't have focus, draw it with the specified colour
				brBack = GetSysColorBrush(COLOR_WINDOW);
				FillRect(dis->hDC, &(dis->rcItem), brBack);
				SetBkColor(dis->hDC, GetSysColor(COLOR_WINDOW));

				// Map lookup current color
				std::map<unsigned int, unsigned int>::iterator pColor = z->colors.find(dis->itemID);
				SetTextColor(
					dis->hDC,
					(pColor == z->colors.end()) ? z->default_color : pColor->second
				);
			}
			/*TextOut(
				dis->hDC,
				dis->rcItem.left + 2, dis->rcItem.top,
				text.c_str(), text.length()
			);*/
			dis->rcItem.left += 2;
			DrawText(
				dis->hDC,
				text.c_str(), text.length(),
				& dis->rcItem,
				DT_NOPREFIX | DT_PATH_ELLIPSIS | DT_EXPANDTABS /* | DT_CENTER*/
			);

			return true; // handled
		} break;

		case WM_MEASUREITEM: {
			MEASUREITEMSTRUCT *mis = (MEASUREITEMSTRUCT*)ev->lParam;
			int multi = z->line_count;
			mis->itemHeight =
				multi * (MGFontManager::getInstance()->getFontPixelHeight() + /*5*/ /*6*/ 5)
			;
			return true;
		} break;

		case WM_KEYUP:
		case WM_CHAR: {
			return z->allow_type_to_select; // maybe chomp
		}

		case WM_KEYDOWN: {
			MGVirtualKeys vk = MGNativeToVK(ev->wParam);
			if (vk != MGVK_NONE) {
				this->onKey.data = MGKC(vk, false, false, false); // FIXME
				this->onKey.invoke();
				
				if (onKey.wasEventDefaultPrevented()) {
					return true; // handled, but this doesn't really
					// prevent the default behaviour unless all other keyboard
					// messages are also chomped
				}
			}
			return z->allow_type_to_select; // maybe chomp
		} break;

	} // switch
	return false; // not otherwise handled
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGProgressBar
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGProgressBar::MGProgressBar(MGWindow* parent, bool isHorizontal, bool isSmooth) {
	this->_controlType = MGCT_PROGRESSBAR;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(PROGRESS_CLASS, 0,
		WS_CHILD | WS_VISIBLE		
		| (isHorizontal ? 0 : PBS_VERTICAL)
		| (isSmooth ? 0 : PBS_SMOOTH),
		0, 0, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0);
	if (! d->hWnd) return;
	SendMessage(d->hWnd, PBM_SETRANGE32, 0, 10000);

	e = new MGEventHandler(this);
	e->subclassWnd(d->hWnd);
}

MGProgressBar::~MGProgressBar() {
	if (e) delete e;
}

long MGProgressBar::_parseEvent(MGEventData *ev) {
	switch(ev->uMsg) {
		case WM_MOUSEMOVE: {
			if (ev->wParam & MK_LBUTTON) {
				onDrag.data = ev->lParam;
				onDrag.invoke();
			}
		} break;
		case WM_LBUTTONDOWN: {
			onMouseDown.data = ev->lParam;
			onMouseDown.invoke();
		} break;
	}
	return false; // message not handled
}

void MGProgressBar::setValue(float fraction) {
	SendMessage(d->hWnd, PBM_SETPOS, (WPARAM)(fraction * 10000.0f), 0);
}

float MGProgressBar::getValue() {
	return (float)SendMessage(d->hWnd, PBM_GETPOS, 0, 0) / 10000.0f;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGLabel
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGLabel::MGLabel(MGWindow* parent, bool alignLeft) {
	this->_controlType = MGCT_LABEL;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(WC_STATIC, 0,
		WS_CHILD | WS_VISIBLE | (alignLeft ? SS_LEFT : SS_CENTER) | SS_PATHELLIPSIS, 
		0, 0, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0);
	if (! d->hWnd) return;

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

void MGLabel::setText(const char* text) {
	SetWindowText(d->hWnd, utf8to16(text).c_str());
}

std::string MGLabel::getText() {
	wchar_t resultBuffer[256] = {0};
	GetWindowText(d->hWnd, resultBuffer, 255);
	return utf16to8(resultBuffer);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGImage
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MGImage::MGImage(MGWindow* parent, int resource_id) {
	this->_controlType = MGCT_IMAGE;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(WC_STATIC, 0,
		WS_CHILD | WS_VISIBLE | SS_BITMAP,
		0, 0, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0);
	if (! d->hWnd) return;

	HBITMAP lImage = (HBITMAP)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(resource_id), IMAGE_BITMAP, 0, 0, 0);
	// optionally could use LR_LOADFROMFILE

	SendMessage(d->hWnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)lImage);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGGroup
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGGroup::MGGroup(MGWindow* parent) {
	this->_controlType = MGCT_GROUP;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(L"BUTTON", L"", 
		WS_CHILD | WS_VISIBLE | BS_GROUPBOX,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0);
	if (! d->hWnd) return;

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

void MGGroup::setText(const char* text) {
	SetWindowText(d->hWnd, utf8to16(text).c_str());
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGComboBox
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGComboBox::MGComboBox(MGWindow* parent, bool bAutoSort) {
	this->_controlType = MGCT_COMBOBOX;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(L"COMBOBOX", L"",
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_TABSTOP | 
		CBS_NOINTEGRALHEIGHT | CBS_DROPDOWNLIST | (bAutoSort ? CBS_SORT : 0),
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0);
	if (! d->hWnd) return;

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

void MGComboBox::allowTextEntry(bool bEnable) { // FIXME
	long style = GetWindowLong(d->hWnd, GWL_STYLE);
	if (bEnable) {
		style &= ~ CBS_DROPDOWNLIST;
		style |= CBS_DROPDOWN;
	} else {
		style &= ~ CBS_DROPDOWN;
		style |= CBS_DROPDOWNLIST;
	}
	SetWindowLong(d->hWnd, GWL_STYLE, style);
}

void MGComboBox::addItem(const char* text) {
	SendMessage(d->hWnd, CB_ADDSTRING, 0, (LPARAM)(utf8to16(text).c_str()));
}

void MGComboBox::clear() {
	SendMessage(d->hWnd, CB_RESETCONTENT, 0, 0);
}

void MGComboBox::setSelected(int idx) {
	SendMessage(d->hWnd, CB_SETCURSEL, idx, 0);
}

int MGComboBox::getSelected() {
	int cursel = (int)SendMessage(d->hWnd, CB_GETCURSEL, 0, 0);
	return (cursel == CB_ERR ? -1 : cursel);
}

std::string MGComboBox::getSelectedText() { //UNTESTED
	wchar_t buff[1024] = {0};
	GetWindowText(d->hWnd, buff, 1023);
	return utf16to8(buff);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGTimer
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGTimer::MGTimer(MGWindow* parent, int interval_ms) {
	this->_controlType = MGCT_TIMER;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	SetTimer(d->ParentHWnd, d->ID, interval_ms, 0);
}

MGTimer::~MGTimer() {
	KillTimer(d->ParentHWnd, d->ID);	
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGHotkey
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGHotkey::MGHotkey(MGWindow* parent, MGKeyCombination kc) {
	this->_controlType = MGCT_HOTKEY;
	//d = new MGControlData;
	parent->_prepareChild(d, this);

	setCombination(kc);
}

MGHotkey::~MGHotkey() {
	UnregisterHotKey(d->ParentHWnd, d->ID);
}

MGKeyCombination MGHotkey::getCombination() {
	return m_kc;
};

void MGHotkey::setCombination(MGKeyCombination kc) {
	if (m_kc.virtkey) {
		UnregisterHotKey(d->ParentHWnd, d->ID);
	}

	if (kc.virtkey) {
		RegisterHotKey(d->ParentHWnd, d->ID,
			(kc.alt ? MOD_ALT : 0) |
			(kc.ctrl ? MOD_CONTROL : 0) |
			(kc.shift ? MOD_SHIFT : 0),
			MGVKToNative(kc.virtkey)
		);
	}

	m_kc = kc;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MGHotkeyControl
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


MGHotkeyControl::MGHotkeyControl(MGWindow* parent) {
	this->_controlType = MGCT_HOTKEYCONTROL;
	parent->_prepareChild(d, this);

	d->hWnd = CreateWindow(HOTKEY_CLASS, L"",
		WS_CHILD | WS_VISIBLE | WS_TABSTOP,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		d->ParentHWnd, (HMENU)d->ID, d->hInstance, 0
	);

	MGFontManager::getInstance()->SetControlFont(d->hWnd);
}

void MGHotkeyControl::setValue(MGKeyCombination val) {
	SendMessage(
		d->hWnd, HKM_SETHOTKEY,
		MAKEWORD(
			MGVKToNative(val.virtkey),
			(val.alt ? HOTKEYF_ALT : 0) |
			(val.ctrl ? HOTKEYF_CONTROL : 0) |
			(val.shift ? HOTKEYF_SHIFT : 0)
		),
		0
	);
}

MGKeyCombination MGHotkeyControl::getValue() {
	MGKeyCombination val;
	long res = SendMessage(d->hWnd, HKM_GETHOTKEY, 0, 0);

	// Decode response
	val.virtkey = MGNativeToVK(LOBYTE(LOWORD(res)));

	char kmod =   HIBYTE(LOWORD(res));
	val.alt =   (kmod & HOTKEYF_ALT    ) ? 1 : 0;
	val.ctrl =  (kmod & HOTKEYF_CONTROL) ? 1 : 0;
	val.shift = (kmod & HOTKEYF_SHIFT  ) ? 1 : 0;
	return val;
}